// |>
// | UCM VIEWER 1-1-1 - UCMV JS [http://www.universalchemicalmarkup.org]
// | Updated: 2014-08-30
// | License: GNU GPL 3 [http://www.gnu.org/licenses/gpl-3.0.html]
// | Author: Jan Mokrý (Main corresponding author:
// |                    Contact information is at the project website.)
// |>

var renderer, scene, sceneAutorotation, sceneRoot, sceneCamera, sceneControls, viewerPath, viewedFilename;
var loadedUcmDefinitionsData, loadedUcmStructuresData;
var loadedRootDescription, loadedRootDescriptionReferences, loadedRootLiteratureReferences;
var currentDescriptionReferences, currentLiteratureReferences, currentPropertyReferences;
var currentStructureIndex, currentStructuresCount;
var quantityDefinitions, unitDefinitions, entryDefinitions, nodeDefinitions, bondDefinitions;
var spriteParameters, lineParameters;

function init(filePath) {
    renderer = new THREE.CanvasRenderer();
    renderer.setSize(window.innerWidth, window.innerHeight);
    $('#CANVAS').append(renderer.domElement);
    scene = new THREE.Scene();
    sceneAutorotation = true;
    sceneRoot = new THREE.Object3D();
    scene.add(sceneRoot);
    sceneCamera = new THREE.PerspectiveCamera(35, window.innerWidth / window.innerHeight, 0.3, 9.5);
    sceneCamera.position.z = 1.65;
    sceneControls = new THREE.TrackballControls(sceneCamera, renderer.domElement);
    sceneControls.addEventListener('change', renderScene);
    viewerPath = $(location).attr('href');

    if (filePath) {
        viewedFilename = filePath.split(/[\\\/]/).pop();
        viewerPath = viewerPath.replace(viewerPath.split(/[\\\/]/).pop(), '');
    } else {
        viewedFilename = viewerPath.split(/[\\\/]/).pop();
        viewerPath = viewerPath.replace(viewedFilename, '');
    }

    loadUcmData();
}

function loadUcmData() {
    $.get(viewerPath + viewedFilename, function(ucmData) {
        loadedUcmDefinitionsData = [];
        loadedUcmStructuresData = [];
        loadedRootDescription = '';
        loadedRootDescriptionReferences = [];
        loadedRootLiteratureReferences = [];
        currentDescriptionReferences = [];
        currentLiteratureReferences = [];
        currentPropertyReferences = [];
        currentStructureIndex = 0;
        quantityDefinitions = {};
        unitDefinitions = {};
        entryDefinitions = {};
        nodeDefinitions = {};
        bondDefinitions = {};

        $(ucmData).find('define').each(function() {
            loadedUcmDefinitionsData.push($(this));
        });

        $(ucmData).find('xi\\:include, include').each(function() {
            var $include = $(this);

            $.ajaxSetup({async: false});

            $.get(viewerPath + $include.attr('href'), function(ucmIncludedData) {
                $(ucmIncludedData).find('define').each(function() {
                    loadedUcmDefinitionsData.push($(this));
                });
            });

            $.ajaxSetup({async: true});
        });

        for (var dataIndex = 0; dataIndex < loadedUcmDefinitionsData.length; dataIndex++) {
            loadedUcmDefinitionsData[dataIndex].find('unitsml\\:Quantity, Quantity').each(
                loadQuantityDefinitions);
            loadedUcmDefinitionsData[dataIndex].find('unitsml\\:Unit, Unit').each(loadUnitDefinitions);
            loadedUcmDefinitionsData[dataIndex].find('bibtexml\\:entry, entry').each(loadEntryDefinitions);
            loadedUcmDefinitionsData[dataIndex].find('node:not([idrefs])').each(loadNodeDefinitions);
        }

        $(ucmData).find('node:not([idrefs])').each(loadNodeDefinitions);
        $(ucmData).find('ucm').children('description').each(loadAndAppendDescription);

        $(ucmData).find('ucm').children('structure').each(function() {
            loadedUcmStructuresData.push($(this));
        });

        currentStructuresCount = loadedUcmStructuresData.length;
        loadStructureData();
    });
}

function loadStructureData() {
    var structure = loadedUcmStructuresData[currentStructureIndex];
    var structureIndex = currentStructureIndex + 1;
    var nextStructureIndex;

    if (structureIndex < currentStructuresCount) {
        nextStructureIndex = structureIndex + 1;
    } else {
        nextStructureIndex = 1;
    }

    if (currentStructuresCount > 1) {
        $('#SCENE-CONTROLS').append('<p id="SCENE-INFO"></p>');
        $('#SCENE-CONTROLS').append('<p><button onclick="loadStructure()">Load Structure '
                                    + nextStructureIndex + '</button></p>');
    }

    currentDescriptionReferences = loadedRootDescriptionReferences.slice(); // Pass array by value.
    currentLiteratureReferences = loadedRootLiteratureReferences.slice(); // Pass array by value.
    currentPropertyReferences = [];
    spriteParameters = {};
    lineParameters = {};
    var structureFormat = structure.attr('format');
    var structureType = structure.attr('type');
    $('#SCENE-INFO').append('Loaded structure ' + structureIndex + '(' + currentStructuresCount + ').');

    if (structureFormat == 'UCM') {
        if (structure.find('node[x][y][z], point').length > 0) {
            $('#SCENE-CONTROLS').append(
                '<p><button onclick="toggleAutorotation()">Toggle Autorotation</button></p>');
            structure.find('node[x][y][z]').each(loadSpriteParameters);
            structure.find('point').each(loadSpriteParameters);
            updateAllSpritesPositions();
            renderAllSprites();
            structure.find('bond').each(loadAndRenderBond);
        } else {
            $('#SCENE-INFO').append(' Visualization is not supported, because this structure does not '
                                    + 'contain any node or point elements with coordinates in 3-dimensional '
                                    + 'space.');
        }
    } else {
        $('#SCENE-INFO').append(' This is a structure ');

        if (structureType == 'STQR') {
            $('#SCENE-INFO').append('query');
        }

        $('#SCENE-INFO').append('recorded using ' + structureFormat + ' format. Thus, visualization is not '
                                + 'supported, but the content of the structure element displayed here can be '
                                + 'copied and processed in some other chemical software: ' + structure.text());
    }

    $('#DESCRIPTIONS > .CONTENT').append('<h3>UCM ROOT</h3>');
    $('#DESCRIPTIONS > .CONTENT').append(loadedRootDescription);

    if (loadedRootLiteratureReferences.length > 0) {
        $('#DESCRIPTIONS > .CONTENT').append('<p>Literature references: '
                                             + loadedRootLiteratureReferences.join(', ') + '</p>');
    }

    structure.find("structure[type='STID']").each(loadAndAppendIdentifier);
    structure.find('description').each(loadAndAppendDescription);
    structure.find("property[type='PR']").each(loadAndAppendProperty);

    if (currentDescriptionReferences.length > 0) {
        $('#DESCRIPTIONS > .CONTENT').append('<h3>REFERENCED DESCRIPTIONS</h3>');
    }

    if (currentPropertyReferences.length > 0) {
        $('#PROPERTIES > .CONTENT').append('<h3>REFERENCED PROPERTIES</h3>');
    }

    for (var dataIndex = 0; dataIndex < loadedUcmDefinitionsData.length; dataIndex++) {
        if (currentDescriptionReferences.length > 0) {
            for (var descriptionId in currentDescriptionReferences) {
                descriptionId = currentDescriptionReferences[descriptionId];
                loadedUcmDefinitionsData[dataIndex].children('#' + descriptionId).each(
                    loadAndAppendDescription);
            }
        }

        if (currentPropertyReferences.length > 0) {
            for (var propertyId in currentPropertyReferences) {
                propertyId = currentPropertyReferences[propertyId];
                loadedUcmDefinitionsData[dataIndex].children('#' + propertyId).each(loadAndAppendProperty);
            }
        }
    }

    appendAllLiteratureReferences();
}

function loadQuantityDefinitions() {
    var $quantity = $(this);
    quantityDefinitions['#' + $quantity.attr('xml:id')] = [
        $quantity.children('unitsml\\:QuantityName').text(),
        $quantity.children('unitsml\\:UnitReference').attr('url')];
}

function loadUnitDefinitions() {
    var $unit = $(this);
    unitDefinitions['#' + $unit.attr('xml:id')] = $unit.children('unitsml\\:UnitName').text();
}

function loadEntryDefinitions() {
    var $entry = $(this);
    var entryId = $entry.attr('id');
    var entryDataItems = {};
    entryDataItems['type'] = $entry.children('*').prop('tagName').split(':')[1];

    $entry.children('*').children('*').each(function() {
        var $entryDataItem = $(this);
        entryDataItems[$entryDataItem.prop('tagName').split(':')[1]] = $entryDataItem.text();
    });

    entryDefinitions[entryId] = entryDataItems;
}

function loadNodeDefinitions() {
    var $node = $(this);
    var nodeId = $node.attr('id');
    var nodeParticles = {};

    $node.children('particle').each(function() {
        var $particle = $(this);
        nodeParticles[$particle.attr('type')] = $particle.attr('counts').replace(' ', ' or ');
    });

    nodeDefinitions[nodeId] = nodeParticles;
}

function loadSpriteParameters() {
    var $element = $(this);
    var elementId = $element.attr('id');
    var elementPosition = new THREE.Vector3(Number($element.attr('x')),
                                            Number($element.attr('y')),
                                            Number($element.attr('z')));
    var spriteRadiusAndMaterial = generateSpriteRadiusAndMaterial($element);
    spriteParameters[elementId] = [elementPosition, spriteRadiusAndMaterial[0], spriteRadiusAndMaterial[1],
                                   $element.attr('idrefs')];
}

function updateAllSpritesPositions() {
    var xCoordinateSum = 0;
    var yCoordinateSum = 0;
    var zCoordinateSum = 0;
    var coordinatesCount = 0;

    for (var elementId in spriteParameters) {
        var elementSpritePosition = spriteParameters[elementId][0];
        xCoordinateSum += elementSpritePosition.x;
        yCoordinateSum += elementSpritePosition.y;
        zCoordinateSum += elementSpritePosition.z;
        coordinatesCount++;
    }

    var centroidCoordinates = new THREE.Vector3(xCoordinateSum / coordinatesCount,
                                                yCoordinateSum / coordinatesCount,
                                                zCoordinateSum / coordinatesCount);

    for (var elementId in spriteParameters) {
        var elementSpritePosition = spriteParameters[elementId][0];
        elementSpritePosition.sub(centroidCoordinates);
    }
}

function renderAllSprites() {
    for (var elementId in spriteParameters) {
        var elementSpriteParameters = spriteParameters[elementId];
        var elementSprite = new THREE.Sprite(elementSpriteParameters[2]);
        var spriteType;
        sceneRoot.add(elementSprite);
        elementSprite.position = elementSpriteParameters[0];
        elementSprite.scale.set(elementSpriteParameters[1] * 2, elementSpriteParameters[1] * 2, 1.0);

        if (elementSpriteParameters[1] == 0.012) {
            spriteType = 'Point';
        } else {
            spriteType = 'Node';
        }

        $('#SCENE-CONTROLS').append('<p><button id="TB-' + elementId + '" onclick="toggleSpriteHighlighting(\''
                                    + elementId + '\')">Highlight ' + spriteType + ' ' + elementId
                                    + '</button></p>');
    }
}

function loadAndRenderBond() {
    var $bond = $(this);
    var bondOrder = $bond.attr('order');

    if ((bondOrder != 'I') && (bondOrder != 'H') && (bondOrder != 'DIP')) {
        var bondId = $bond.attr('id');
        var bondLineMaterials = generateLineMaterials(bondOrder);
        bondDefinitions[bondId] = [bondOrder, []];
        lineParameters[bondId] = [bondLineMaterials];
        $('#SCENE-CONTROLS').append('<p><button id="TB-' + bondId + '" onclick="toggleLineHighlighting(\''
                                    + bondId + '\')">Highlight Bond ' + bondId + '</button></p>');

        $bond.children('particle').each(function() {
            var $particle = $(this);
            bondDefinitions[bondId][1].push([$particle.attr('idrefs'), $particle.attr('counts')]);
        });

        if ($bond.is('[idrefs]')) {
            var bondIdrefs = $bond.attr('idrefs').split(' ');
            var bondLineGeometry = new THREE.Geometry();
            var bondLineCoordinates = findLineCoordinates(
                spriteParameters[bondIdrefs[0]][1],
                spriteParameters[bondIdrefs[1]][1],
                spriteParameters[bondIdrefs[0]][0],
                spriteParameters[bondIdrefs[1]][0]);
            bondLineGeometry.vertices.push(bondLineCoordinates[0]);
            bondLineGeometry.vertices.push(bondLineCoordinates[1]);
            renderBondLine(bondLineGeometry, bondLineMaterials);
        } else {
            $bond.children('join').each(function() {
                var $join = $(this);
                var bondIdrefs = $join.attr('idrefs').split(' ');

                if ($join.text() == 'SQ') {
                    loadAndRenderSequentialBond(bondIdrefs, bondLineMaterials);
                } else if ($join.text() == 'CC') {
                    loadAndRenderCyclicBond(bondIdrefs, bondLineMaterials);
                } else if ($join.text() == 'CT') {
                    loadAndRenderCenteredBond(bondIdrefs, bondLineMaterials);
                }
            });
        }
    }
}

function loadAndRenderSequentialBond(bondIdrefs, bondLineMaterials) {
    for (var idrefsIndex = 0; idrefsIndex < (bondIdrefs.length - 1); idrefsIndex++) {
        var bondLineGeometry = new THREE.Geometry();
        var bondLineCoordinates = findLineCoordinates(
            spriteParameters[bondIdrefs[idrefsIndex]][1],
            spriteParameters[bondIdrefs[idrefsIndex + 1]][1],
            spriteParameters[bondIdrefs[idrefsIndex]][0],
            spriteParameters[bondIdrefs[idrefsIndex + 1]][0]);
        bondLineGeometry.vertices.push(bondLineCoordinates[0]);
        bondLineGeometry.vertices.push(bondLineCoordinates[1]);
        renderBondLine(bondLineGeometry, bondLineMaterials);
    }
}

function loadAndRenderCyclicBond(bondIdrefs, bondLineMaterials) {
    for (var idrefsIndex = 0; idrefsIndex < bondIdrefs.length; idrefsIndex++) {
        var bondLineGeometry = new THREE.Geometry();
        var bondLineCoordinates;

        if (idrefsIndex < (bondIdrefs.length - 1)) {
            bondLineCoordinates = findLineCoordinates(
                spriteParameters[bondIdrefs[idrefsIndex]][1],
                spriteParameters[bondIdrefs[idrefsIndex + 1]][1],
                spriteParameters[bondIdrefs[idrefsIndex]][0],
                spriteParameters[bondIdrefs[idrefsIndex + 1]][0]);
        } else {
            bondLineCoordinates = findLineCoordinates(
                spriteParameters[bondIdrefs[idrefsIndex]][1],
                spriteParameters[bondIdrefs[0]][1],
                spriteParameters[bondIdrefs[idrefsIndex]][0],
                spriteParameters[bondIdrefs[0]][0]);
        }

        bondLineGeometry.vertices.push(bondLineCoordinates[0]);
        bondLineGeometry.vertices.push(bondLineCoordinates[1]);
        renderBondLine(bondLineGeometry, bondLineMaterials);
    }
}

function loadAndRenderCenteredBond(bondIdrefs, bondLineMaterials) {
    for (var idrefsIndex = 1; idrefsIndex < bondIdrefs.length; idrefsIndex++) {
        var bondLineGeometry = new THREE.Geometry();
        var bondLineCoordinates = findLineCoordinates(
            spriteParameters[bondIdrefs[0]][1],
            spriteParameters[bondIdrefs[idrefsIndex]][1],
            spriteParameters[bondIdrefs[0]][0],
            spriteParameters[bondIdrefs[idrefsIndex]][0]);
        bondLineGeometry.vertices.push(bondLineCoordinates[0]);
        bondLineGeometry.vertices.push(bondLineCoordinates[1]);
        renderBondLine(bondLineGeometry, bondLineMaterials);
    }
}

function renderBondLine(bondLineGeometry, bondLineMaterials) {
    for (var materialIndex = 0; materialIndex < bondLineMaterials.length; materialIndex++) {
        sceneRoot.add(new THREE.Line(bondLineGeometry, bondLineMaterials[materialIndex]));
    }
}

function generateSpriteRadiusAndMaterial(element) {
    var spriteRadius;
    var spriteMaterial;
    var protonsCount;

    if (element.is('[idrefs]')) {
        protonsCount = nodeDefinitions[element.attr('idrefs')]['P'];
    } else {
        var elementId = element.attr('id');

        if (elementId in nodeDefinitions) {
            protonsCount = nodeDefinitions[elementId]['P'];
        } else {
            protonsCount = 0;
        }
    }

    // Sprite radius is used for relative scaling of node sprites to approximately 2/3 of appropriate covalent
    // radius from "http://en.wikipedia.org/wiki/Covalent_radius", while point sprites are given a small fixed
    // radius.
    if (protonsCount == 0) {
        spriteRadius = 0.02;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureDefault.png')});
    } else if (protonsCount == 1) {
        spriteRadius = 0.031;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureH.png')});
    } else if (protonsCount == 5) {
        spriteRadius = 0.084;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureB.png')});
    } else if (protonsCount == 6) {
        spriteRadius = 0.076;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureC.png')});
    } else if (protonsCount == 7) {
        spriteRadius = 0.071;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureN.png')});
    } else if (protonsCount == 8) {
        spriteRadius = 0.066;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureO.png')});
    } else if (protonsCount == 17) {
        spriteRadius = 0.102;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureCl.png')});
    } else if (protonsCount == 26) {
        spriteRadius = 0.132;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureFe.png')});
    } else if (protonsCount == 27) {
        spriteRadius = 0.126;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureCo.png')});
    } else if (protonsCount == 35) {
        spriteRadius = 0.12;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureBr.png')});
    } else if (protonsCount == 78) {
        spriteRadius = 0.136;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTexturePt.png')});
    } else {
        spriteRadius = 0.1;
        spriteMaterial = new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureDefault.png')});
    }

    return [spriteRadius * 0.6, spriteMaterial];
}

function generateLineMaterials(bondOrder) {
    var lineMaterials = [];

    if (bondOrder == 'S') {
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 2, linecap: 'butt'}));
    } else if (bondOrder == 'D') {
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 8, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 4, linecap: 'butt'}));
    } else if (bondOrder == 'T') {
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 14, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 10, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 2, linecap: 'butt'}));
    } else if (bondOrder == 'Q') {
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 20, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 16, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({linewidth: 8, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 4, linecap: 'butt'}));
    } else if (bondOrder == 'PS') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 2, linecap: 'butt'}));
    } else if (bondOrder == 'PD') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 8, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 4, linecap: 'butt'}));
    } else if (bondOrder == 'PT') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 14, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 10, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 2, linecap: 'butt'}));
    } else if (bondOrder == 'PQ') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 20, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 16, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AAAA, linewidth: 8, linecap: 'butt'}));
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x000000, linewidth: 4, linecap: 'butt'}));
    } else if (bondOrder == 'A') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0x00AA00, linewidth: 5, linecap: 'butt'}));
    } else if (bondOrder == 'DL') {
        lineMaterials.push(new THREE.LineBasicMaterial({color: 0xAA0000, linewidth: 5, linecap: 'butt'}));
    }

    return lineMaterials;
}

function findLineCoordinates(distanceFromStart, distanceFromEnd, startCoordinates, endCoordinates) {
    var lineUnitVector = new THREE.Vector3();
    var lineStartVector = new THREE.Vector3();
    var lineEndVector = new THREE.Vector3();
    var lineStartCoordinates = new THREE.Vector3();
    var lineEndCoordinates = new THREE.Vector3();
    lineUnitVector.subVectors(startCoordinates, endCoordinates).normalize();
    lineStartVector.copy(lineUnitVector).multiplyScalar(distanceFromStart);
    lineEndVector.copy(lineUnitVector).multiplyScalar(distanceFromEnd);
    lineStartCoordinates.subVectors(startCoordinates, lineStartVector);
    lineEndCoordinates.addVectors(endCoordinates, lineEndVector);
    return [lineStartCoordinates, lineEndCoordinates];
}

function loadAndAppendDescription() {
    var $description = $(this);
    var parentElement = $description.parent();
    var parentElementName = parentElement.prop('tagName');
    var saveRootDescription = false;
    var descriptionTitle = '';

    if (parentElementName == 'define') {
        descriptionId = $description.attr('id');
        descriptionTitle += ('<h4 id="' + descriptionId + '">' + descriptionId
                             + ' (EXTERNAL DESCRIPTION)</h4>');
    } else if (parentElementName == 'ucm') {
        saveRootDescription = true;
    } else if ((parentElementName == 'particle') || (parentElementName == 'stereo')) {
        var grandparentElement = parentElement.parent();
        descriptionTitle += (grandparentElement.prop('tagName').toUpperCase() + '[@id='
                             + grandparentElement.attr('id') + ']/' + parentElementName.toUpperCase() + '['
                             + (grandparentElement.children(parentElementName).index(parentElement) + 1)
                             + ']');
    } else {
        descriptionTitle += (parentElementName.toUpperCase() + '[@id=' + parentElement.attr('id') + ']');
    }

    if (saveRootDescription) {
        loadedRootDescription = $description.contents();
    } else {
        $('#DESCRIPTIONS > .CONTENT').append('<h3>' + descriptionTitle + '</h3>');
        // Append cloned description contents to avoid removing contents from original description element.
        $('#DESCRIPTIONS > .CONTENT').append($description.contents().clone());
    }

    if ($description.is('[idrefs]')) {
        var descriptionReferences = $description.attr('idrefs').split(' ');

        for (var referenceIndex = 0; referenceIndex < descriptionReferences.length; referenceIndex++) {
            var descriptionReference = descriptionReferences[referenceIndex];
            descriptionReferences[referenceIndex] = ('<a href="#' + descriptionReference + '">'
                                                     + descriptionReference + '</a>');

            if (saveRootDescription && (loadedRootDescriptionReferences.indexOf(descriptionReference) < 0)) {
                loadedRootDescriptionReferences.push(descriptionReference);
            }

            if (currentDescriptionReferences.indexOf(descriptionReference) < 0) {
                currentDescriptionReferences.push(descriptionReference);
            }
        }

        if (!saveRootDescription) {
            $('#DESCRIPTIONS > .CONTENT').append('<p>Description references: '
                                                 + descriptionReferences.join(', ') + '</p>');
        }
    }

    if ($description.is('[litrefs]')) {
        var literatureReferences = $description.attr('litrefs').split(' ');

        for (var referenceIndex = 0; referenceIndex < literatureReferences.length; referenceIndex++) {
            var literatureReference = literatureReferences[referenceIndex];
            literatureReferences[referenceIndex] = ('<a href="#' + literatureReference + '">'
                                                    + literatureReference + '</a>');

            if (saveRootDescription && (loadedRootLiteratureReferences.indexOf(literatureReference) < 0)) {
                loadedRootLiteratureReferences.push(literatureReference);
            }

            if (currentLiteratureReferences.indexOf(literatureReference) < 0) {
                currentLiteratureReferences.push(literatureReference);
            }
        }

        if (!saveRootDescription) {
            $('#DESCRIPTIONS > .CONTENT').append('<p>Literature references: '
                                                 + literatureReferences.join(', ') + '</p>');
        }
    }
}

function appendAllLiteratureReferences() {
    if ((getAssociativeArraySize(entryDefinitions) != 0) && (currentLiteratureReferences.length > 0)) {
        $('#DESCRIPTIONS > .CONTENT').append('<h3>LITERATURE REFERENCES</h3>');

        for (var entryId in currentLiteratureReferences) {
            entryId = currentLiteratureReferences[entryId];
            var entryDataItems = entryDefinitions[entryId];
            var entryData = '<p>';

            for (var entryDataItemName in entryDataItems) {
                var entryDataItem = entryDataItems[entryDataItemName];

                if (entryDataItemName == 'author') {
                    entryData += ('<b>' + entryDataItemName.toUpperCase() + ':</b> '
                                  + entryDataItem.split(' and ').join('; ') + '<br/>');
                } else if (entryDataItemName == 'url') {
                    entryData += ('<b>' + entryDataItemName.toUpperCase() + ':</b> '
                                  + '<a href="' + entryDataItem + '">' + entryDataItem + '</a><br/>');
                } else if (entryDataItemName != 'type'){
                    entryData += ('<b>' + entryDataItemName.toUpperCase() + ':</b> '
                                  + entryDataItem + '<br/>');
                }
            }

            entryData += '</p>';
            $('#DESCRIPTIONS > .CONTENT').append('<h4 id="' + entryId + '">' + entryId
                                                 + ' (' + entryDataItems['type'].toUpperCase() + ')</h4>');
            $('#DESCRIPTIONS > .CONTENT').append(entryData);
        }
    }
}

function loadAndAppendIdentifier() {
    var $identifier = $(this);
    var identifierFormat = $identifier.attr('format');
    identifierFormat = identifierFormat.replace('S-INCHI', 'STANDARD-INCHI');
    identifierFormat = identifierFormat.replace('-U', '-(UNIQUE)');
    identifierFormat = identifierFormat.replace(/-/g, ' ');
    $('#IDENTIFIERS > .CONTENT').append('<h3>' + identifierFormat + '</h3>');
    $('#IDENTIFIERS > .CONTENT').append('<p>' + $identifier.text() + '</p>');
}

function loadAndAppendProperty() {
    var $property = $(this);

    if ($property.is('[idrefs]')) {
        var propertyIdrefs = $property.attr('idrefs');
        $('#PROPERTIES > .CONTENT').append('<p>Property reference: <a href="#' + propertyIdrefs + '">'
                                           + propertyIdrefs + '</a></p>');

        if (currentPropertyReferences.indexOf(propertyIdrefs) < 0) {
            currentPropertyReferences.push(propertyIdrefs);
        }
    } else {
        var propertyId = $property.attr('id');
        var propertyType = $property.attr('type');
        var propertyQuantity = quantityDefinitions[$property.attr('quantity')];
        var propertyQuantityName = propertyQuantity[0];
        var propertyUnitName = unitDefinitions[propertyQuantity[1]];
        var propertyValues = $property.children('values').text().split(' ');
        var propertyTitle = '';

        if (propertyType == 'CN') {
            propertyTitle += ('<h4 id="' + propertyId + '">' + propertyQuantityName.toUpperCase()
                              + ' [CONDITION @id=' + propertyId + ']</h4>');
        } else if (propertyType == 'ER') {
            propertyTitle += ('<h4 id="' + propertyId + '">' + propertyQuantityName.toUpperCase()
                              + ' [ERROR @id=' + propertyId + ']</h4>');
        } else if (propertyType == 'PR')  {
            propertyTitle += ('<h3 id="' + propertyId + '">' + propertyQuantityName.toUpperCase()
                              +  ' [@id=' + propertyId + ']</h3>');
        }

        $('#PROPERTIES > .CONTENT').append(propertyTitle);
        $('#PROPERTIES > .CONTENT').append('<p>Values [' + propertyUnitName + ']:<br/>'
                                           + propertyValues.join('<br/>') + '</p>');
        $property.children('property').each(loadAndAppendProperty);
    }
}

function getAssociativeArraySize(associativeArray) {
    var size = 0;

    for (var key in associativeArray) {
        if (associativeArray.hasOwnProperty(key)) {
            size++;
        }
    }

    return size;
}

function animateScene() {
    requestAnimationFrame(animateScene);
    sceneControls.update();

    if (sceneAutorotation) {
        var rotationAmount = Date.now() * 0.0002;
        sceneRoot.rotation.x = rotationAmount;
        sceneRoot.rotation.y = rotationAmount * 0.6;
    }

    renderScene();
}

function renderScene() {
    renderer.render(scene, sceneCamera);
}

function loadStructure() {
    if (currentStructureIndex + 1 < currentStructuresCount) {
        currentStructureIndex++;
    } else {
        currentStructureIndex = 0;
    }

    scene.remove(sceneRoot);
    sceneRoot = new THREE.Object3D();
    scene.add(sceneRoot);
    $('#DESCRIPTIONS > .CONTENT').empty().append('<h2>DESCRIPTIONS</h2>');
    $('#IDENTIFIERS > .CONTENT').empty().append('<h2>IDENTIFIERS</h2>');
    $('#PROPERTIES > .CONTENT').empty().append('<h2>PROPERTIES</h2>');
    $('#SCENE-CONTROLS').empty();
    $('#HIGHLIGHTED-ITEMS-DETAILS h3, #HIGHLIGHTED-ITEMS-DETAILS p').remove();
    $('#HIGHLIGHTED-ITEMS-DETAILS').hide();
    loadStructureData();
}

function loadFile() {
    adaptInterfaceToResolution();
    $('#CANVAS').empty();
    $('#DESCRIPTIONS > .CONTENT').empty().append('<h2>DESCRIPTIONS</h2>');
    $('#IDENTIFIERS > .CONTENT').empty().append('<h2>IDENTIFIERS</h2>');
    $('#PROPERTIES > .CONTENT').empty().append('<h2>PROPERTIES</h2>');
    $('#SCENE-CONTROLS').empty();
    $('#HIGHLIGHTED-ITEMS-DETAILS h3, #HIGHLIGHTED-ITEMS-DETAILS p').remove();
    $('#HIGHLIGHTED-ITEMS-DETAILS').hide();

    if ($("input[name='FILE']").length > 0) {
        init($("input[name='FILE']").val());
    } else {
        init('');
    }

    animateScene();
}

function toggleAutorotation() {
    if (sceneAutorotation) {
        sceneAutorotation = false;
    } else {
        sceneAutorotation = true;
    }
}

function toggleSpriteHighlighting(idString) {
    var elementSpriteParameters = spriteParameters[idString];

    if (elementSpriteParameters.length == 4) {
        var elementDefinitionId = elementSpriteParameters[3];

        if (elementDefinitionId === undefined) {
            elementDefinitionId = idString;
        }

        if (elementDefinitionId in nodeDefinitions) {
            var nodeParticles = nodeDefinitions[elementDefinitionId];
            var nodeDetails = '<b>Particle counts:</b><br/>';

            for (paticleType in nodeParticles) {
                nodeDetails += nodeParticles[paticleType];

                if (paticleType == 'P') {
                    nodeDetails += ' Proton(s)<br/>';
                } else if (paticleType == 'N') {
                    nodeDetails += ' Neutron(s)<br/>';
                } else if (paticleType == 'E') {
                    nodeDetails += ' Electron(s)<br/>';
                } else if (paticleType == 'BE') {
                    nodeDetails += ' Bonding electron(s)<br/>';
                } else if (paticleType == 'NBE') {
                    nodeDetails += ' Non-bonding electron(s)<br/>';
                }
            }

            nodeDetails += ('<b>Coordinates:</b><br/>' + 'X: ' + elementSpriteParameters[0].x + '<br/>'
                            + 'Y: ' + elementSpriteParameters[0].y + '<br/>'
                            + 'Z: ' + elementSpriteParameters[0].z);
            $('#HIGHLIGHTED-ITEMS-DETAILS').append('<h3 class="' + idString + '">NODE ' + idString
                                                   + ' DETAILS</h3><p class="' + idString + '">'
                                                   + nodeDetails + '</p>');
        } else {
            var pointDetails = ('<b>Coordinates:</b><br/>' + 'X: ' + elementSpriteParameters[0].x + '<br/>'
                            + 'Y: ' + elementSpriteParameters[0].y + '<br/>'
                            + 'Z: ' + elementSpriteParameters[0].z);
            $('#HIGHLIGHTED-ITEMS-DETAILS').append('<h3 class="' + idString + '">POINT ' + idString
                                                   + ' DETAILS</h3><p class="' + idString + '">'
                                                   + pointDetails + '</p>');
        }

        var elementHighlightSprite = new THREE.Sprite(new THREE.SpriteMaterial(
            {map: THREE.ImageUtils.loadTexture('UcmvTextureHighlight.png')}));
        sceneRoot.add(elementHighlightSprite);
        elementHighlightSprite.position = elementSpriteParameters[0];
        elementHighlightSprite.scale.set(elementSpriteParameters[1] * 2, elementSpriteParameters[1] * 2, 1.0);
        spriteParameters[idString][4] = elementHighlightSprite;
        $('#HIGHLIGHTED-ITEMS-DETAILS').show();
    } else {
        sceneRoot.remove(spriteParameters[idString][4]);
        spriteParameters[idString].splice(4, 1);
        $('#HIGHLIGHTED-ITEMS-DETAILS .' + idString).remove();
        $('#HIGHLIGHTED-ITEMS-DETAILS').not(':has(h3, p)').hide();
    }

    toggleHighlightButton(idString);
}

function toggleLineHighlighting(idString) {
    var bondLineParameters = lineParameters[idString];

    if (bondLineParameters.length == 1) {
        var lineWidthMultiplier = 3;
        var bondOrder = bondDefinitions[idString][0];
        var bondParticles = bondDefinitions[idString][1];
        var bondDetails = '';

        if (bondParticles.length > 0) {
            bondDetails += '<b>Particle counts:</b><br/>';

            for (var particleIndex = 0; particleIndex < bondParticles.length; particleIndex++) {
                var particleOrigin;
                particleOriginId = bondParticles[particleIndex][0];

                if (particleOriginId in bondDefinitions) {
                    particleOrigin = 'shared with bond ';
                } else if (particleOriginId in spriteParameters) {
                    particleOrigin = 'from node ';
                } else {
                    particleOrigin = 'from structure ';
                }

                bondDetails += (bondParticles[particleIndex][1] + ' Bonding electron(s) ' + particleOrigin
                                + particleOriginId + '<br/>')
            }
        }

        if (bondOrder == 'S') {
            bondOrder = 'Single bond';
        } else if (bondOrder == 'D') {
            bondOrder = 'Double bond';
        } else if (bondOrder == 'T') {
            bondOrder = 'Triple bond';
        } else if (bondOrder == 'Q') {
            bondOrder = 'Quadruple bond';
        } else if (bondOrder == 'PS') {
            bondOrder = 'Partial single bond';
        } else if (bondOrder == 'PD') {
            bondOrder = 'Partial double bond';
        } else if (bondOrder == 'PT') {
            bondOrder = 'Partial triple bond';
        } else if (bondOrder == 'PQ') {
            bondOrder = 'Partial quadruple bond';
        } else if (bondOrder == 'A') {
            bondOrder = 'Aromatic bond';
        } else if (bondOrder == 'DL') {
            bondOrder = 'Delocalized or other bond';
        }

        bondDetails += ('<b>Order:</b><br/>' + bondOrder);
        $('#HIGHLIGHTED-ITEMS-DETAILS').append('<h3 class="' + idString + '">BOND ' + idString
                                               + ' DETAILS</h3><p class="' + idString + '">'
                                                   + bondDetails + '</p>');

        for (var materialIndex = 0; materialIndex < bondLineParameters[0].length; materialIndex++) {
            bondLineParameters[0][materialIndex].linewidth *= lineWidthMultiplier;
        }

        lineParameters[idString][1] = lineWidthMultiplier;
        $('#HIGHLIGHTED-ITEMS-DETAILS').show();
    } else {
        for (var materialIndex = 0; materialIndex < bondLineParameters[0].length; materialIndex++) {
            bondLineParameters[0][materialIndex].linewidth /= lineParameters[idString][1];
        }

        lineParameters[idString].splice(1, 1);
        $('#HIGHLIGHTED-ITEMS-DETAILS .' + idString).remove();
        $('#HIGHLIGHTED-ITEMS-DETAILS').not(':has(h3, p)').hide();
    }

    toggleHighlightButton(idString);
}

function toggleHighlightButton(idString) {
    if ($('#TB-' + idString).text().indexOf('Highlight') == 0) {
        $('#TB-' + idString).css({'color': '#0066AA'});
        $('#TB-' + idString).text($('#TB-' + idString).text().replace('Highlight', 'Unhighlight'));
    } else {
        $('#TB-' + idString).css({'color': '#000'});
        $('#TB-' + idString).text($('#TB-' + idString).text().replace('Unhighlight', 'Highlight'));
    }
}

function toggleDivAndButton(idString) {
    $('#' + idString).toggle();

    if ($('#' + idString).css('display') == 'none') {
        $('#TB-' + idString).css({'color': '#000'});
    } else {
        $('#TB-' + idString).css({'color': '#0066AA'});
    }

    resetOtherDivsAndButtons(idString);
}

function resetOtherDivsAndButtons(idString) {
    var allIdStrings = ['CONTROLS', 'DESCRIPTIONS', 'IDENTIFIERS', 'PROPERTIES', 'HINTS'];

    for (var idStringIndex = 0; idStringIndex < allIdStrings.length; idStringIndex++) {
        var currentIdString = allIdStrings[idStringIndex];

        if (currentIdString != idString) {
            $('#' + currentIdString).hide();
            $('#TB-' + currentIdString).css({'color': '#000'});
        }
    }
}

function windowResize() {
    adaptInterfaceToResolution();

    if (!($('#CANVAS').is(':empty'))) {
        sceneCamera.aspect = window.innerWidth / window.innerHeight;
        sceneCamera.updateProjectionMatrix();
        renderer.setSize(window.innerWidth, window.innerHeight);
        renderScene();
        sceneControls = new THREE.TrackballControls(sceneCamera, renderer.domElement);
        sceneControls.addEventListener('change', renderScene);
    }
}

function adaptInterfaceToResolution() {
    if ((window.innerWidth < 480) || (window.innerHeight < 190)) {
        $('#ERRORS').not(':has(p.ERROR-1)').append(
            '<p class="ERROR-1"><b>Error:</b> Display window is too small (minimal size is 480x190 pixels). '
                + 'Resize the web browser window or switch to a display with higher resolution.</p>');
        $('#ERRORS').show();
    } else {
        $('p.ERROR-1').remove();
        $('#ERRORS').hide();
    }

    if (window.innerWidth < 675) {
        $('h1').css({'color': '#FFF'});
    } else {
        $('h1').css({'color': '#00BBBB'});
    }
}

console.log('UCM VIEWER 1-1-1');
adaptInterfaceToResolution();
window.addEventListener('resize', windowResize, false);
